@@ -15,7 +15,7 @@
 #        Tristan Van Berkom <tristan.vanberkom@codethink.co.uk>
 #        Tiago Gomes <tiago.gomes@codethink.co.uk>
 
-from typing import TYPE_CHECKING, Optional, Dict, Union, List
+from typing import TYPE_CHECKING, Optional, Dict, Union, List, Sequence
 
 import os
 import urllib.parse
@@ -31,18 +31,21 @@ from ._exceptions import LoadError
 from .exceptions import LoadErrorReason
 from ._options import OptionPool
 from .node import ScalarNode, MappingNode, ProvenanceInformation, _assert_symbol_name
-from ._pluginfactory import ElementFactory, SourceFactory, load_plugin_origin
-from .types import CoreWarnings, _HostMount, _SourceMirror, _SourceUriPolicy
+from ._pluginfactory import ElementFactory, SourceFactory, SourceMirrorFactory, load_plugin_origin
+from .types import CoreWarnings, _HostMount, _SourceUriPolicy
 from ._projectrefs import ProjectRefs, ProjectRefStorage
 from ._loader import Loader, LoadContext
 from .element import Element
 from ._includes import Includes
 from ._workspaces import WORKSPACE_PROJECT_FILE
 from ._remotespec import RemoteSpec
+from .sourcemirror import SourceMirror
+from .source import SourceError
 
 
 if TYPE_CHECKING:
     from ._context import Context
+    from .plugins.elements.junction import JunctionElement
 
 
 # Project Configuration file
@@ -56,7 +59,7 @@ class ProjectConfig:
         self.base_variables = {}  # The base set of variables
         self.element_overrides = {}  # Element specific configurations
         self.source_overrides = {}  # Source specific configurations
-        self.mirrors = {}  # Dictionary of _SourceAlias objects
+        self.mirrors = {}  # Dictionary of SourceMirror objects
         self.default_mirror = None  # The name of the preferred mirror.
         self._aliases = None  # Aliases dictionary
 
@@ -81,7 +84,7 @@ class Project:
         directory: Optional[str],
         context: "Context",
         *,
-        junction: Optional[object] = None,
+        junction: Optional["JunctionElement"] = None,
         cli_options: Optional[Dict[str, str]] = None,
         default_mirror: Optional[str] = None,
         parent_loader: Optional[Loader] = None,
@@ -97,11 +100,12 @@ class Project:
 
         self.load_context: LoadContext  # The LoadContext
         self.loader: Optional[Loader] = None  # The loader associated to this project
-        self.junction: Optional[object] = junction  # The junction Element object, if this is a subproject
+        self.junction: Optional["JunctionElement"] = junction  # The junction Element object, if this is a subproject
 
         self.ref_storage: Optional[ProjectRefStorage] = None  # Where to store source refs
         self.refs: Optional[ProjectRefs] = None
         self.junction_refs: Optional[ProjectRefs] = None
+        self.disallow_subproject_uris: bool = False
 
         self.config: ProjectConfig = ProjectConfig()
         self.first_pass_config: ProjectConfig = ProjectConfig()
@@ -115,6 +119,7 @@ class Project:
 
         self.element_factory: Optional[ElementFactory] = None  # ElementFactory for loading elements
         self.source_factory: Optional[SourceFactory] = None  # SourceFactory for loading sources
+        self.source_mirror_factory: Optional[SourceMirrorFactory] = None  # SourceMirrorFactory
 
         self.sandbox: Optional[MappingNode] = None
         self.splits: Optional[MappingNode] = None
@@ -195,6 +200,28 @@ class Project:
     #                     Public Methods                   #
     ########################################################
 
+    # get_alias_url():
+    #
+    # Fetch the value of a URL alias
+    #
+    # Args:
+    #    alias: The alias
+    #    first_pass: Whether to use first pass configuration (for junctions)
+    #
+    # Returns:
+    #    The alias substitution
+    #
+    def get_alias_url(self, alias: str, *, first_pass: bool = False) -> Optional[str]:
+        if first_pass:
+            config = self.first_pass_config
+        else:
+            config = self.config
+
+        if config._aliases:
+            return config._aliases.get_str(alias, default=None)
+
+        return None
+
     # translate_url():
     #
     # Translates the given url which may be specified with an alias
@@ -210,16 +237,28 @@ class Project:
     # This method is provided for :class:`.Source` objects to resolve
     # fully qualified urls based on the shorthand which is allowed
     # to be specified in the YAML
-    def translate_url(self, url, *, first_pass=False):
-        if first_pass:
-            config = self.first_pass_config
-        else:
-            config = self.config
+    def translate_url(self, url, *, source, first_pass=False):
 
         if url and utils._ALIAS_SEPARATOR in url:
             url_alias, url_body = url.split(utils._ALIAS_SEPARATOR, 1)
-            alias_url = config._aliases.get_str(url_alias, default=None)
+            alias_url = self.get_alias_url(url_alias, first_pass=first_pass)
             if alias_url:
+                if self.junction:
+                    parent_project = self.junction._get_project()
+                    parent_alias = self.junction.get_parent_alias(url_alias)
+                    if parent_alias:
+                        # Delegate translation to parent project
+                        return parent_project.translate_url(
+                            parent_alias + utils._ALIAS_SEPARATOR + url_body, source=source, first_pass=first_pass
+                        )
+                    elif parent_project.disallow_subproject_uris:
+                        raise SourceError(
+                            "{}: Parent project did not provide a mapping for alias '{}' and disallowed usage of unmapped aliases".format(
+                                source, url_alias
+                            ),
+                            reason="missing-alias-mapping",
+                        )
+
                 url = alias_url + url_body
 
         return url
@@ -371,23 +410,51 @@ class Project:
     # Returns:
     #    bool: Whether the alias is declared in the scope of this project
     #
-    def alias_exists(self, alias, *, first_pass=False):
+    def alias_exists(self, alias, *, source, first_pass=False):
         if first_pass:
             config = self.first_pass_config
         else:
             config = self.config
 
+        if self.junction:
+            parent_project = self.junction._get_project()
+            parent_alias = self.junction.get_parent_alias(alias)
+            if parent_alias:
+                if parent_project.alias_exists(parent_alias, source=source, first_pass=first_pass):
+                    return True
+                else:
+                    raise SourceError(
+                        "{}: Mapped alias '{}' for subproject alias '{}' is invalid in the parent project".format(
+                            self.junction, parent_alias, alias
+                        ),
+                        reason="invalid-source-alias",
+                    )
+            elif parent_project.disallow_subproject_uris:
+                raise SourceError(
+                    "{}: Parent project did not provide a mapping for alias '{}' and disallowed usage of unmapped aliases".format(
+                        source, alias
+                    ),
+                    reason="missing-alias-mapping",
+                )
+
         return config._aliases.get_str(alias, default=None) is not None
 
     # get_alias_uris()
     #
     # Args:
-    #    alias (str): The alias.
-    #    first_pass (bool): Whether to use first pass configuration (for junctions)
-    #    tracking (bool): Whether we want the aliases for tracking (otherwise assume fetching)
+    #    alias: The alias.
+    #    first_pass: Whether to use first pass configuration (for junctions)
+    #    tracking: Whether we want the aliases for tracking (otherwise assume fetching)
     #
-    # Returns a list of every URI to replace an alias with
-    def get_alias_uris(self, alias, *, first_pass=False, tracking=False):
+    # Returns:
+    #    A list of alias substitutions found. Each substitution is either a string
+    #    (to be prepended to the rest of the URL instead of the alias) or a
+    #    SourceMirror object.
+    #
+    def get_alias_uris(
+        self, alias: str, *, first_pass: bool = False, tracking: bool = False
+    ) -> Sequence[Union[SourceMirror, str, None]]:
+
         if first_pass:
             config = self.first_pass_config
         else:
@@ -396,18 +463,28 @@ class Project:
         if not alias or alias not in config._aliases:  # pylint: disable=unsupported-membership-test
             return [None]
 
-        uri_list = []
+        if self.junction:
+            parent_project = self.junction._get_project()
+            parent_alias = self.junction.get_parent_alias(alias)
+            if parent_alias:
+                # Delegate translation to parent project
+                return parent_project.get_alias_uris(parent_alias, first_pass=first_pass, tracking=tracking)
+            elif parent_project.disallow_subproject_uris:
+                return [None]
+
+        uri_list: List[Union[SourceMirror, str]] = []
         policy = self._context.track_source if tracking else self._context.fetch_source
 
         if policy in (_SourceUriPolicy.ALL, _SourceUriPolicy.MIRRORS) or (
             policy == _SourceUriPolicy.USER and self._mirror_override
         ):
             for mirror_name, mirror in config.mirrors.items():
-                if alias in mirror.aliases:
-                    if mirror_name == config.default_mirror:
-                        uri_list = mirror.aliases[alias] + uri_list
-                    else:
-                        uri_list += mirror.aliases[alias]
+                list_to_add = mirror._get_alias_uris(alias)
+
+                if mirror_name == config.default_mirror:
+                    uri_list = list_to_add + uri_list
+                else:
+                    uri_list += list_to_add
 
         if policy in (_SourceUriPolicy.ALL, _SourceUriPolicy.ALIASES):
             uri_list.append(config._aliases.get_str(alias))
@@ -786,7 +863,19 @@ class Project:
 
         # Junction configuration
         junctions_node = pre_config_node.get_mapping("junctions", default={})
-        junctions_node.validate_keys(["duplicates", "internal"])
+        junctions_node.validate_keys(["duplicates", "internal", "disallow-subproject-uris"])
+
+        if self.junction and self.junction._get_project().disallow_subproject_uris:
+            # If the parent project doesn't allow subproject URIs, this must
+            # be enforced for nested subprojects as well.
+            self.disallow_subproject_uris = True
+
+            # The `disallow-subproject-uris` flag also implies fatal `unaliased-url` in subprojects
+            # to ensure no subproject URIs escape the parent project's control.
+            if CoreWarnings.UNALIASED_URL not in self._fatal_warnings:
+                self._fatal_warnings.append(CoreWarnings.UNALIASED_URL)
+        else:
+            self.disallow_subproject_uris = junctions_node.get_bool("disallow-subproject-uris", default=False)
 
         # Parse duplicates
         junction_duplicates = junctions_node.get_mapping("duplicates", default={})
@@ -1025,9 +1114,9 @@ class Project:
         # even if the mirrors are specified in user configuration.
         variables.expand(mirrors_node)
 
-        # Collect _SourceMirror objects
+        # Collect SourceMirror objects
         for mirror_node in mirrors_node:
-            mirror = _SourceMirror.new_from_node(mirror_node)
+            mirror = self.source_mirror_factory.create(self._context, self, mirror_node)
             output.mirrors[mirror.name] = mirror
             if not output.default_mirror:
                 output.default_mirror = mirror.name
@@ -1087,6 +1176,7 @@ class Project:
         pluginbase = PluginBase(package="buildstream.plugins")
         self.element_factory = ElementFactory(pluginbase)
         self.source_factory = SourceFactory(pluginbase)
+        self.source_mirror_factory = SourceMirrorFactory(pluginbase)
 
         # Load the plugin origins and register them to their factories
         origins = config.get_sequence("plugins", default=[])
@@ -1096,6 +1186,8 @@ class Project:
                 self.element_factory.register_plugin_origin(kind, origin, conf.allow_deprecated)
             for kind, conf in origin.sources.items():
                 self.source_factory.register_plugin_origin(kind, origin, conf.allow_deprecated)
+            for kind, conf in origin.source_mirrors.items():
+                self.source_mirror_factory.register_plugin_origin(kind, origin, conf.allow_deprecated)
 
     # _warning_is_fatal():
     #
